---------------------------------------------------------------------------
---------------------------------------------------------------------------
--                                                                       --
--    Curry Offline Partial Evaluator: annotation procedure              --
--                                                                       --
--    Version September 2006                                                        --
--                                                                       --
--    Authors:                                                           --
--    Gustavo Arroyo, J. Guadalupe Ramos, Josep Silva, German Vidal      --
--    (Technical University of Valencia)                                 --
--                                                                       --
---------------------------------------------------------------------------
---------------------------------------------------------------------------

module Annotate(annot) where

---------------------------------------------------------------------------
-- Imports
---------------------------------------------------------------------------

import AbstractCurry             --analysis is performed at the level of 
import AbstractCurryPrinter      --source Curry (rather than at the level
                                 --of Flat Curry

import schgMono                  --size-change analysis
import btaAbst                   --simple monovariant BTA analysis

import Unsafe


---------------------------------------------------------------------------
-- Data Types
---------------------------------------------------------------------------

----------------- size change graph types ---------------------------------
--type Fname  = String                --- function name
--type Arity  = Int                   --- function arity
--type Src    = Int                   --- source node
--type Tgt    = Int                   --- target node
--data Order  = GrT | Equ            
--type Edge   = (Src, Tgt, Order)
--type BTD    = [(Fname,Arity,BTEnv)] --- BindingTime Division 

------------------------f---------------g----------arches------------------
--data   Graph = G (Fname,Arity) (Fname,Arity) [Edge] 
--data  MGraph = MG Graph [Graph]
---------------------------------------------------------------------------

---------------------------------------------------------------------------
-- Auxiliary Functions
---------------------------------------------------------------------------

writeCurryFile :: String -> CurryProg -> IO()
writeCurryFile file prog = writeFile file (showProg prog)

---------------------------------------------------------------------------
-- Main Functions
---------------------------------------------------------------------------

annot file =
   do 
     putStrLn "Offline Narrowing-Driven Partial Evaluator"
     putStrLn "(Version 0.1 of November 2006)"
     putStrLn "(Technical University of Valencia)\n"
     putStrLn "Pre-processing analysis and annotation of source program"
     putStrLn "by using size-change graphs and binding-time analysis\n"
     putStrLn "Loading AbstractCurry program"
     progAbs <- readCurry (file)
     putStrLn "Computing size-change graphs"
     let gr = getGraphs progAbs
     putStrLn "Computing binding-time division"
     let btD = getd progAbs [D,D]
     --putStrLn (show btD)
     --putStrLn "Program loaded"
     --putStrLn "Annotating program & linearizing..."
     let Aprog = anotateProg progAbs gr btD
     --putStrLn "Program succesfully annotated"
     --putStrLn "Transforming annotated program to Curry..."
     writeCurryFile (file++"_ann.curry") Aprog
     putStrLn ("Annotated program stored in <<"++file++"_ann.curry>>")

getd :: CurryProg -> BTEnv -> BTD
getd p tau0 = zipWith makeFinD gas div
         where  (gas,div) = explore p tau0
 

-- 

anotateProg :: CurryProg -> [MGraph]-> BTD -> CurryProg
anotateProg (CurryProg name mods types functions operations) graphs btD = 
  (CurryProg nameA mods types newfuns operations)
      where afunctions = (anotateFuns nameA functions graphs btD)
            linAnFuncs = map (linearAP btD) afunctions
            newfuns = map (changeModName name) linAnFuncs
            nameA = name++"_ann"

changeModName :: String -> CFuncDecl -> CFuncDecl
changeModName mod (CFunc name ar vis ftype (CExternal fname)) =
   CFunc name ar vis ftype (CExternal fname)
changeModName mod (CFunc name ar vis ftype (CRules feval rules)) =
  (CFunc name ar vis ftype (CRules feval (map (changeModRule mod) rules)))

changeModRule mod (CRule pats guards locals) =
  CRule pats (map (changeModGuards mod) guards) 
             (map (changeModLocal mod) locals)

changeModGuards mod (e1,e2) = (changeModExpr mod e1, changeModExpr mod e2)

changeModLocal mod (CLocalFunc fundec) = CLocalFunc (changeModName mod fundec)
changeModLocal mod (CLocalPat lpat exp locals) =
  CLocalPat lpat (changeModExpr mod exp) (map (changeModLocal mod) locals)
changeModLocal mod (CLocalVar lvar) = CLocalVar lvar

changeModExpr _ (CVar i) = CVar i
changeModExpr _ (CLit l) = CLit l
changeModExpr mod (CSymbol (fmod,fname)) = CSymbol (nmod,fname)
  where
    nmod = if mod==fmod then fmod++"_ann" else fmod
changeModExpr mod (CApply e1 e2) = CApply (changeModExpr mod e1)
                                          (changeModExpr mod e2)
changeModExpr mod (CLambda pats e) = CLambda pats (changeModExpr mod e)
changeModExpr mod (CLetDecl locals e) =
  CLetDecl (map (changeModLocal mod) locals) (changeModExpr mod e)
changeModExpr mod (CDoExpr statements) =
  CDoExpr (map (changeModSt mod) statements)
changeModExpr mod (CListComp e statements) =
  CListComp (changeModExpr mod e) (map (changeModSt mod) statements)
changeModExpr mod (CCase e branches) =
  CCase (changeModExpr mod e) (map (changeModBranch mod) branches)


changeModSt mod (CSExpr e) = CSExpr (changeModExpr mod e)
changeModSt mod (CSPat pat e) = CSPat pat (changeModExpr mod e)
changeModSt mod (CSLet locals) = CSLet (map (changeModLocal mod) locals)

changeModBranch mod (CBranch pat e) = CBranch pat (changeModExpr mod e)

-- if (itsgen linAnFuncs)
--    then (CurryProg nameA mods types (linAnFuncs ++ [gen]) operations)
--    else (CurryProg nameA mods types linAnFuncs operations)
-- where
--    gen      = (CFunc (nameA,"GEN") 1 (AbstractCurry.Public) (CFuncType (CTVar (0,"a")) (CTVar (0,"a"))) (CRules CFlex [(CRule [(CPVar (0,"x"))] [((CSymbol ("Prelude","success")),(CVar (0,"x")))] [])]))
          

anotateFuns :: String -> [CFuncDecl] -> [MGraph] -> BTD -> [CFuncDecl]
anotateFuns mod functions graph btD = map (anotateFun mod graph btD) functions


-- This function anotates a single function of an AbstractCurry program

anotateFun :: String -> [MGraph] -> BTD -> CFuncDecl -> CFuncDecl
anotateFun mod graphs btD (CFunc name arity vis typ rules) =
                         (CFunc name arity vis typ Arules)
    where Arules = anotateRules mod rules graphs btD
--          btE    = srchbtE name BTD


-- This function anotates the rules of a function

anotateRules :: String -> CRules -> [MGraph] -> BTD -> CRules
anotateRules mod (CRules ev rules) graphs btD = (CRules ev (map (anotateRule mod graphs btD) rules))
anotateRules _ (CExternal name) _ _ = (CExternal name)


-- This function anotates a single rule of a function definition

anotateRule :: String -> [MGraph] -> BTD -> CRule -> CRule
anotateRule mod graphs btD (CRule patterns rhs localDecls) = (CRule patterns Arhs localDecls)
    where Arhs = map (anotateRhs mod graphs btD) rhs


-- This function anotates the right hand sides of a rule

anotateRhs :: String -> [MGraph] -> BTD -> (CExpr,CExpr) -> (CExpr,CExpr)
anotateRhs mod graphs btD (guard,expr) = (guard,Aexpr)
    where Aexpr = anotateExpr mod graphs btD expr


-- This function anotates an expression

anotateExpr :: String -> [MGraph] -> BTD -> CExpr -> CExpr
anotateExpr _ _ _ (CVar name) =  (CVar name)
anotateExpr _ _ _ (CLit lit)  =  (CLit lit)
anotateExpr _ _ _ (CSymbol name) = (CSymbol name)
anotateExpr mod graphs btD (CApply expr1 expr2) = anotateApp mod graphs btD (CApply expr1 expr2)
anotateExpr _ _ _ (CLambda patterns expr) = (CLambda patterns expr)   -- Lambda Abstractions not treated in this version
anotateExpr mod graphs btD (CLetDecl lDecls expr) = (CLetDecl (anotateDecls mod graphs btD lDecls) (anotateExpr mod graphs btD expr))
anotateExpr mod graphs btD (CDoExpr statements) = (CDoExpr (anotateSts mod graphs btD statements))
anotateExpr mod graphs btD (CListComp expr statements) = (CListComp (anotateExpr mod graphs btD expr) (anotateSts mod graphs btD statements))
anotateExpr mod graphs btD (CCase expr branches) = (CCase (anotateExpr mod graphs btD expr) (anotateBrs mod graphs btD branches))


-- This function anotates a list of local declarations

anotateDecls :: String -> [MGraph] -> BTD -> [CLocalDecl] -> [CLocalDecl]
anotateDecls mod graphs btD lDecls = map (anotateDecl mod graphs btD) lDecls


-- This function anotates a single local declaration


anotateDecl :: String -> [MGraph] -> BTD -> CLocalDecl -> CLocalDecl
anotateDecl mod graphs btD (CLocalFunc funcDecl) = (CLocalFunc (anotateFun mod graphs btD funcDecl)) -- local function declaration
anotateDecl mod graphs btD (CLocalPat pat expr lDecls) = (CLocalPat pat (anotateExpr mod graphs btD expr) (anotateDecls mod graphs btD lDecls)) -- local pattern declaration
anotateDecl _ _ _ (CLocalVar var) = (CLocalVar var) -- local free variable declaration


-- This function anotates a list of statements

anotateSts :: String -> [MGraph] -> BTD -> [CStatement] -> [CStatement]
anotateSts mod graphs btD sts = map (anotateSt mod graphs btD) sts


-- This function anotates a single statement

anotateSt :: String -> [MGraph] -> BTD -> CStatement -> CStatement
anotateSt mod graphs btD (CSExpr expr) = (CSExpr (anotateExpr mod graphs btD expr))       -- an expression (I/O action or boolean)
anotateSt mod graphs btD (CSPat pat expr) = (CSPat pat (anotateExpr mod graphs btD expr)) -- a pattern definition
anotateSt mod graphs btD (CSLet lDecls) = (CSLet (anotateDecls mod graphs btD lDecls))    -- a local let declaration


-- This function anotates a list of branches

anotateBrs :: String -> [MGraph] -> BTD -> [CBranchExpr] -> [CBranchExpr]
anotateBrs mod graphs btD brs = map (anotateBr mod graphs btD) brs


-- This function anotates a single branch

anotateBr :: String -> [MGraph] -> BTD -> CBranchExpr -> CBranchExpr
anotateBr mod graphs btD (CBranch pat expr) = (CBranch pat (anotateExpr mod graphs btD expr))


-- This function anotates an application

anotateApp :: String -> [MGraph] -> BTD -> CExpr -> CExpr                                 --decia solo args, puse      
anotateApp mod graphs btD (CApply expr1 expr2) = if isCiclic                              --*** (reverse args)
                                                    then  genApp mod graphs btD (CApply expr1 expr2) (reverse args)
                                                    else (CApply (anotateExpr mod graphs btD expr1) (anotateExpr mod graphs btD expr2))
    where (isCiclic, args) = (isInGraph graphs btD (CApply expr1 expr2))


-- This function generalizes the specified arguments of an application (a function call)

genApp :: String -> [MGraph] -> BTD -> CExpr -> [Bool] -> CExpr 
genApp mod graphs btD (CApply expr1 expr2) [arg] = 
           if (arg == True)
              then (CApply expr1 (CApply (CSymbol (mod,"GEN")) (anotateExpr mod graphs btD expr2))) 
              else (CApply expr1 (anotateExpr mod graphs btD expr2))
genApp mod graphs btD (CApply expr1 expr2) (arg1:arg2:args) =   -- Expr1 must be a "CApply"
       if (arg1 == True)
          then (CApply (genApp mod graphs btD expr1 (arg2:args)) (CApply (CSymbol (mod,"GEN")) (anotateExpr mod graphs btD expr2))) 
          else (CApply (genApp mod graphs btD expr1 (arg2:args)) (anotateExpr mod graphs btD expr2)) 
 

-- This function takes a function and checks if this function belongs to a cicle (with the graphs). If so, it returns True
-- if not, it returns False. In addition, it returns a list of booleans indicating for each argument wether it should be 
-- generalized or not.
-- Here it should be applied also the bindingTime values, The condition is: No anotar si el multigrafo contiene al menos un arco
-- etiquetado con orden estricto y estático.

isInGraph :: [MGraph] -> BTD -> CExpr -> (Bool, [Bool])
isInGraph graphs btD (CApply expr1 expr2) = if (fun=="NotInGraphs")
                                               then (False, [])
                                               else (True, (computeArgs graphs fun btD (take arity (repeat False))))                     
      where (fun,arity) = findFun graphs (CApply expr1 expr2)
--             initArgs n = if (n==0) then [] else (1:(initArgs (n-1)))


-- This function checks if a function belongs to a cicle (with the graphs). If so, it returns the name of the function and its arity
-- if not, it returns "NotInGraphs".

findFun :: [MGraph] -> CExpr -> (String, Int)
findFun graphs (CApply expr1 _) = case expr1 of
                                      (CApply expr2 expr3) -> findFun graphs (CApply expr2 expr3)
                                      (CSymbol (_,fun))    -> searchFunGs graphs fun 
                                      _                    -> ("NotInGraphs",0)


-- This function checks if a function belongs to a cicle (with the graphs). If so, it returns the name of the function and its arity
-- if not, it returns "NotInGraphs".

searchFunGs :: [MGraph] -> String -> (String, Int)
searchFunGs [] _ = ("NotInGraphs",0)
searchFunGs (g:graphs) fun = if (name=="NotInGraph")
                             then searchFunGs graphs fun 
                             else (name,arity)
      where (name,arity) = searchFunG g fun


-- This function checks if a function belongs to a cicle (with the graphs). If so, it returns the name of the function and its arity
-- if not, it returns "NotInGraph".

searchFunG :: MGraph -> String -> (String, Int)
searchFunG (MG (G (name,arity) _ _) _) fun = if (name==fun)
                                         then (name,arity)
                                         else ("NotInGraph",0)


-- This function returns a list of booleans indicating for each argument of a given function wether it should be 
-- generalized or not.


tst_computeArgs = computeArgs [Gr2,Gr1] "ack" [("acker",1,[D]),("ack",2,[D,D])] [False,False] 


Gr1 = (MG (G ("ack",2) ("ack",2) [(1,1,GrT)] )          [(G ("ack",2) ("ack",2) [(1,1,GrT)])])
Gr2 = (MG (G ("ack",2) ("ack",2) [(2,2,GrT),(1,1,Equ)]) [(G ("ack",2) ("ack",2) [(2,2,GrT),(1,1,Equ)])])
--[("acker",1,[D]),("ack",2,[D,D])]

                                                           --- ***una funcion puede tener mas de dos grafos maxims
computeArgs :: [MGraph] -> String -> BTD -> [Bool] -> [Bool]
computeArgs [] _ _ args = args
computeArgs (graph:graphs) fun btD args = if (name=="NotInGraph")
                                  then computeArgs graphs fun btD args
                                  else let ars= computegraph graph args
                                       in  computeArgs graphs fun btD ars
          where (name,_) = searchFunG graph fun
---------------------------------------------------------------------
-- Se agrego para procesar (mg@((MG (G (_,arity) _ edges) _ )):graphs)
-- ya que no lo soporta el compìlador pakcs 1.6.0
-------------------------------------------------------------------- gad18ago06
computegraph (MG (G (_,arity) _ edges) _ ) args = combineArgs args (extractArgs edges (take arity (repeat True)))
                                                                        
------------------------f---------------g----------arches------------------
--data   Graph = G (Fname,Arity) (Fname,Arity) [Edge] 
--data  MGraph = MG Graph [Graph]
---------------------------------------------------------------------------

                                                           --- ***una funcion puede tener mas de dos grafos maxims
--computeArgs :: [MGraph] -> String -> [Bool] -> [Bool]
--computeArgs [] _ args = args
--computeArgs (mg@((MG (G (_,arity) _ edges) _ )):graphs) fun args = if (name=="NotInGraph")
--                                  then computeArgs graphs fun args
--                                  else let ars= combineArgs args (extractArgs edges (take arity (repeat True)))
--                                       in  computeArgs graphs fun ars
--      where (name,_) = searchFunG mg fun



-- This function computes the list of arguments that should be generalized by applying an OR to two list of arguments
-- True means "generalize"; False means "do not generalize"

combineArgs :: [Bool] -> [Bool] -> [Bool]
combineArgs [] [] = []
combineArgs (a:as) (b:bs) = ((a || b):(combineArgs as bs))


-- This function computes the list of arguments that should be generalized from a graph
-- True means "generalize"; False means "do not generalize"

extractArgs :: [Edge] -> [Bool] -> [Bool] 
extractArgs [] args = args
extractArgs ((src, _, _):edges) args = extractArgs edges (markArg src args)


-- This function marks an argument of a list as not generalizable

markArg :: Int -> [Bool] -> [Bool] 
markArg n (arg:args) = if (n==1)
                           then (False:args)
                           else (arg:(markArg (n-1) args))


linearAP :: BTD -> CFuncDecl -> CFuncDecl
linearAP btD (CFunc (mn,fn) ar CVi CTy cRules) = if (srchBTEfun btD fn) == []
                                                    then (CFunc (mn,fn) ar CVi CTy cRules)
                                                    else (CFunc (mn,fn) ar CVi CTy (linearCRs mn ar btEnv cRules))
         where btEnv = srchBTEfun btD fn

--type BTD         = [(Fname,Arity,BTEnv)]    --- BindingTime Division   
srchBTEfun :: BTD -> String -> BTEnv
srchBTEfun []                _   = []
srchBTEfun ((fn,_,btE):divs) cfn = if fn == cfn then btE
                                                else srchBTEfun divs cfn

linearCRs ::String -> Int -> BTEnv -> CRules -> CRules 
linearCRs _  _ _     (CExternal enam)   = (CExternal enam)
linearCRs mn _ btEnv (CRules cEv rules) = (CRules cEv (map (linearCR mn btEnv) rules))


linearCR :: String ->  BTEnv -> CRule -> CRule 
linearCR mn  btE (CRule lcpat [(lofg,cexp)] lCDcl) =
      case lCDcl of
           []                     -> (CRule lcpat [(lofg,linExpr)] [])
           [(CLocalPat cp ce [])] -> (CRule lcpat [(lofg,linExpr)] [(CLocalPat cp (linearexp mn lcpatbtE ce iltd) [])])
           [(CLocalVar cvarin)]   -> (CRule lcpat [(lofg,linExpr)] [(CLocalVar cvarin)])
   where linExpr  =  linearexp mn lcpatbtE cexp iltd 
         lcpatbtE =  createbtEindx lcpat btE
         iltd     =  length (flatlcpat lcpat) < length (flatcexp cexp) || (leneachvar (flatlcpat lcpat) (flatcexp cexp))



-- si el no. de variables en la parte izquierda es menor que 
-- el correspondiente de la derecha; entonces lineariza la parte derecha de la regla
-- de otra forma deja la parte derecha como esta. O si hay más de una ocurrencia de 
-- una variable en la parte derecha entonces lineariza la parte derecha de la regla

linearexp mn lcpatbtE cexp vlhltvrh = if vlhltvrh
                                           then (linearCexp mn cexp lcpatbtE)  else cexp

--intoPatt
-- por el momento no linearizara: CDoExpr (realizar analisis), CLambda
linearCexp mn cexp lcpatbtE = 
   case cexp of
       (CCase (CVar cvi) bs) -> (CCase (CVar cvi) (map (intoBr mn lcpatbtE) bs))
       (CDoExpr lstat)       -> (CDoExpr lstat)
       (CLambda lcpat ce)    -> (CLambda lcpat ce)   
       _                     -> let (_,expr) = linearizeExp mn (flatcexpw cexp) cexp lcpatbtE in expr

intoBr mn lcpatbtE (CBranch cpattern ce) = (CBranch cpattern (linearCexp mn ce lcpatbtE))
  
 
--linearizara parcialmente: CLetDecl,CListComp
linearizeExp mn fcexp e lcpatbtE =
   case e of
      (CApply _ _)          -> (processApply mn fcexp e lcpatbtE)
      (CLetDecl cloc exp)   -> ((fst (linearizeExp mn fcexp exp lcpatbtE)),(CLetDecl cloc (snd (linearizeExp mn fcexp exp lcpatbtE)) ))
      (CListComp exp lstat) -> ((fst (linearizeExp mn fcexp exp lcpatbtE)),(CListComp (snd (linearizeExp mn fcexp exp lcpatbtE)) lstat))
      _                     -> (fcexp,e)


-- Si la (cvin) de la variable en cuestion, esta en el flatcexp entonces remover del flatcexp esa variable y 
-- devolver al Cexpresion (CVar cvin) con la flatcexpr recortada. En caso que la (cvin) de variable en
-- cuestion no se encuentre en la flatcexp de la regla actual; entonces sera una variable repetida y 
-- susceptible de ser generalizada

linearize mn cvin fcexp lcpatbtE = 
     case fcexp of
        []                     ->  if (dinamic cvin lcpatbtE)
                                      then (fcexp,(CApply (CSymbol (mn,"GEN")) (CVar cvin)))
                                      else (fcexp,(CVar cvin))
        _                      ->  if (elem (CVar cvin) fcexp)
                                      then ((remove [] (CVar cvin) fcexp),(CVar cvin))
                                      else if (dinamic cvin lcpatbtE)
                                             then (fcexp,(CApply (CSymbol (mn,"GEN")) (CVar cvin)))
                                             else (fcexp,(CVar cvin))


processApply mn fcexp e lcpatbtE = 
    case e of 
       (CLit _)           -> (fcexp,e)
       (CVar cvin)        -> linearize mn cvin fcexp lcpatbtE
       (CSymbol _)        -> (fcexp,e)
       (CApply lce rce)   -> let (nfc,tlce)= processApply mn fcexp lce lcpatbtE  in (fst (processApply mn nfc rce lcpatbtE),(CApply tlce (snd (processApply mn nfc rce lcpatbtE))))



--processApply :: String -> [CExpr] -> CExpr -> [(CPattern,BindingTime)] -> ([CExpr],CExpr)
--processApply _  fcexp (CLit clit)  _ = (fcexp,(CLit clit))
--processApply _  fcexp (CSymbol qn) _ = (fcexp,(CSymbol qn))
--processApply mn fcexp (CVar cvin) lcpatbtE  =  linearize mn cvin fcexp lcpatbtE
--processApply mn fcexp (CApply lce rce) lcpatbtE = let (nfc,tlce)= processApply mn fcexp lce lcpatbtE
--                                                    in (fst (processApply mn nfc rce lcpatbtE),(CApply tlce (snd (processApply mn nfc rce lcpatbtE)) ))



remove acum  _           []                   = acum
remove acum (CVar cvin) ((CVar cvinl):rfcexp) = if cvin == cvinl then (acum++rfcexp)
                                                   else remove ((CVar cvinl):acum) (CVar cvin) rfcexp 

createbtEindx :: [CPattern]-> BTEnv ->  [(CPattern,BindingTime)]
createbtEindx lcpat btE = zip lcpat btE

-- Si no encuentra la CVarIName en [(CPattern,BindingTime)] entonces es una variable dinamica
-- pero eso lo debe hacer el btaAbst (BTA del programa AbstractCurry), por lo tanto dejaremos 
-- este caso en False
dinamic :: CVarIName -> [(CPattern,BindingTime)] -> Bool
dinamic _    []                 = False
dinamic cvin (cpatbtE:rcpatbtE) = if (cvincpatbtE cvin cpatbtE) then True
                                                                 else (dinamic cvin rcpatbtE)
 

cvincpatbtE cvin ((CPVar vin),bind)       = if (vin==cvin) && (bind == D) then True else False 
cvincpatbtE cvin ((CPComb _ lcpat), bind) = foldl (||) False (map (cvcpatbtE cvin bind) lcpat)


cvcpatbtE cvin bind (CPVar vin)      = if (vin==cvin) && (bind == D) then True else False 
cvcpatbtE cvin bind (CPComb _ lcpat) =  foldl (||) False (map (cvcpatbtE cvin bind) lcpat)

flatcexp :: CExpr -> [CExpr]
flatcexp cexp =
   case cexp of
      (CVar cvin)          -> (CVar cvin):[]
      (CLit _)             -> []
      (CSymbol _ )         -> []
      (CApply le re)       -> (flatcexp le)++(flatcexp re)
      (CLetDecl _ ce)      -> (flatcexp ce)
      (CDoExpr lstat)      -> (flatlstat  lstat)
      (CListComp ce lstat) -> (flatcexp ce)++(flatlstat lstat)
      (CCase ce lbrex)     -> (flatcexp ce)++(flatlbrex lbrex)

flatlstat []           = []
flatlstat (stat:rstat) =  (flatstat stat)++(flatlstat rstat)

flatstat (CSExpr cexp)  = flatcexp cexp
flatstat (CSPat _ cexp) = flatcexp cexp
flatstat (CSLet lcdec)  = foldr (++) [] (map (flatletwhere ) lcdec)


flatletwhere clocDcl =
    case clocDcl of 
        (CLocalPat _ cexpr _)  -> flatcexp cexpr         ---where
        (CLocalVar _ )         -> []
        (CLocalFunc _)         -> []

flatlbrex []      = []
flatlbrex (brex:rbrex) = (flatbrex brex)++(flatlbrex rbrex)

flatbrex (CBranch _  cexp)   = flatcexp cexp



itsgen ::  [CFuncDecl] ->Bool
itsgen rAfuns = foldl (||) False (map (searchgenCrs ) rAfuns) 


searchgenCrs :: CFuncDecl -> Bool
searchgenCrs (CFunc _ _ _ _ (CExternal _)) = False
searchgenCrs (CFunc _ _ _ _ (CRules _ lrules)) = foldl (||) False (map (searchlCr ) lrules)


searchlCr :: CRule -> Bool
searchlCr (CRule _ ltupreg _) = foldl (||) False (map (searchinExpr ) ltupreg)


searchinExpr :: (CExpr,CExpr) -> Bool
searchinExpr (_,expr) = (searchExpr expr) 


searchExpr :: CExpr -> Bool
searchExpr (CVar _) = False
searchExpr (CLit _) = False
searchExpr (CSymbol (_,name)) = if (name == "GEN") then True else False
searchExpr (CApply e1 e2) = (searchExpr e1) || (searchExpr e2)
searchExpr (CLambda _ _) = False
searchExpr (CLetDecl lCldecl e) = (searchlcdec lCldecl) || (searchExpr e)
searchExpr (CDoExpr lCstatm)   = searchlstatm lCstatm
searchExpr (CListComp e lCstatm) = (searchExpr e) || (searchlstatm lCstatm)
searchExpr (CCase e lbranchexp) = (searchExpr e) || (searchlbranchex lbranchexp)
--searchExpr (CLambda lcpattern e) = (searchExpr e) || (searchlcpatt lcpattern) 

searchlcdec :: [CLocalDecl] -> Bool
searchlcdec rclocd = foldl (||) False (map (searchcdec ) rclocd) 


searchcdec :: CLocalDecl -> Bool
searchcdec (CLocalVar _) = False
searchcdec (CLocalFunc cfuncdec) = searchgenCrs cfuncdec
searchcdec (CLocalPat _ e lclocald) = (searchExpr e) || (searchlcdec lclocald)



searchlstatm :: [CStatement] -> Bool
searchlstatm rstatm = foldl (||) False (map (searchstatm ) rstatm)


searchstatm :: CStatement -> Bool
searchstatm (CSExpr e) = searchExpr e
searchstatm (CSPat _ e) = (searchExpr e)
searchstatm (CSLet lClocdec) = searchlcdec lClocdec
--searchstatm (CSPat cpatt e) = (searchcpatt cpatt) || (searchExpr e)

searchlbranchex :: [CBranchExpr] -> Bool
searchlbranchex rbrex = foldl (||) False (map (searchbrex ) rbrex)


searchbrex :: CBranchExpr -> Bool
searchbrex (CBranch _ e) = (searchExpr e)

--srchbtE:: name BTD 
srchbtE (_,fn)  []                   = (fn,0,[S])  
srchbtE (mn,fn) ((fname,a,btE):rBTD )= if fn == fname then (fname,a,btE)
                                                      else srchbtE (mn,fn) rBTD

leneachvar []                    _       = False
leneachvar ((CPVar cvin):rflcpat) fcexp  = if ncvin > 1 then True
                                             else leneachvar rflcpat fcexp
                            where ncvin = countcvin cvin fcexp 0

countcvin _    []                    count = count
countcvin cvin ((CVar ccvin):rfcexp) count = if cvin == ccvin then countcvin cvin rfcexp count+1
                                                              else countcvin cvin rfcexp count

