------------------------------------------------------------------------------
--- This is the main module of the offline partial evaluator
---
--- It provides three facilities:
--- a) simpleAnnotate "program", which implements a simple analysis
---    based on a syntactic characterization of quasi-terminating
---    program such that those terms that violate this characterization
---    are annotated (returns "program_sann.curry")
--- b) annotate "program", which implements a quasi-termination analysis
---    based on size-change graphs and a simple monovariant BTA in order
---    to annotate the source program (returns "program_ann.curry")
--- c) mix "program_ann", which implements the specialization phase of 
---    the partial evaluator
--- 
--- @author Gustavo Arroyo, JGuadalupe Ramos, Josep Silva, German Vidal
--- @version September 2006
------------------------------------------------------------------------------

module OffPeval(mix) where

import FlatCurry          --pakcs distribution (for metaprogramming)
import Unsafe             --pakcs distribution (for debugging)

import annot(annR)                                                --gad11ago06
import Annotate(annot)    --size change graphs annotation module  --gad22ago06

import OffPretty(ppExpr,ppTree,ppPairs,ppFormatList)
                          --pretty printing
import GNNTrees(GNNTree(..),mark_memo)    
                          --data structure for gnn trees
import General(hasGEN,generalize,delGEN,delGEN2)
                          --generalization
import RLNT(unfold,rlnt,propBinding)
                          --implements the RLNT calculus
import Resultants(extractResultants,buildRen,renResultants)
                          --extraction of resultants
import PostUnf(postUnfolding)
                          --post-unfolding simplification
import Prog2Flat(resultant2fundecl,fundecl2prog,fundecl2progA)
                          --builds the new FlatCurry program
import Utils(extractMainCallLhs,deleteRenamings,funsIn,isCons,isConsRooted,
             decons,isOpRooted,extractBranches,existsRen,isCaseVar,maxVarIndex,extractMainCallRhs)
                          --several utilities


--the following flag determines whether the specialization is added to
--the original program (true, e.g., for benchmarking run times) 
-- or replaces it (false, e.g., for measuring code size reduction)

mixFlag = True 

-- should some debug information be printed during partial evaluation?
pevalDebug = False        --shows the set of partially evaluated terms (and memos)
pevalDebugUnfold = False  --shows every unfolded term and its result

--this flags are used to determine the information shown by the system:
gnnFlag     = False  --should the GNN tree be shown?
gnnMemoFlag = False  --should the GNN tree with memoized nodes be shown?
resulFlag   = False  --should resultants be shown?
renFlag     = False  --should the computed renaming be shown?
renresFlag  = False  --should the renamed resultants be shown?
postFlag    = False  --should the program after postunfolding be shown?

printVersion =
  do putStrLn "Offline Narrowing-Driven Partial Evaluator"
     putStrLn "(Version 0.1 of November 2006)"
     putStrLn "(Technical University of Valencia)\n"

printVersion1 =
  do putStrLn "Offline Narrowing-Driven Partial Evaluator"
     putStrLn "(Version 0.1 of November 2006)"
     putStrLn "(Technical University of Valencia)\n"
     putStrLn "Pre-processing analysis and annotation of source program\n"

printVersion2 =
  do putStrLn "Offline Narrowing-Driven Partial Evaluator"
     putStrLn "(Version 0.1 of November 2006)"
     putStrLn "(Technical University of Valencia)\n"
     putStrLn "Specialization from annotated program\n"

simpleAnnotate :: String -> IO ()
simpleAnnotate p =
    do printVersion1
       prog <- readFlatCurry p
       let progAnn = annR prog 
       putStrLn $ "\n Writing annotated program in \"" ++ file ++ "\"...\n"   
       writeFCY file progAnn
    where file = p++"_sann.fcy"

--annotate just calls to annot in module Annotate
annotate p = annot p

-------------------------------------------------------------
-- Se adapto la función mix, ya que ahora tomara directamente
-- un programa en formato de FlatCurry (i.e.,".fcy") en lugar 
-- de tomar un programa en código fuente Curry
-------------------------------------------------------------
--     prog <- readFlatCurry p                                --gad11ago06
mix0 :: String -> IO()
mix0 p  = 
  do printVersion
     let pr = p ++ ".fcy"
     prog <- readFlatCurryFile pr
     let call =  extractMainCallLhs prog  --extractMainCallRhs prog
     mx p call prog

-------------------------------------------------------------
-- Se creo la función mix, para que hubiera la opción de tomar
-- un programa en formato de Curry (i.e.,".curry") para probar 
-- programa anotados con an anotador de Size Change Graphs
-------------------------------------------------------------
--     prog <- readFlatCurry p                                --gad18ago06
mix :: String -> IO()
mix p  = 
  do printVersion
     prog <- readFlatCurry p
     let call =  extractMainCallLhs prog  --extractMainCallRhs prog
     mx p call prog


mx :: String -> Expr -> Prog -> IO ()
mx p e prog = 
  do putStrLn ("\nMain call: "++(ppExpr 0 e)++"\n")
     if gnnFlag then putStrLn ("\nGNN tree: \n"++(ppTree 0 gr)++"\n")            else done
     if gnnMemoFlag then putStrLn ("\nGNN tree (with memo nodes marked): \n"  ++(ppTree 0 grm)++"\n")
                    else done
     if resulFlag then putStrLn ("\nResultants: \n"  ++(ppPairs resultants "=")++"\n")
                  else done
     if renFlag then putStrLn ("\nRenaming: \n"++(ppPairs renaming "->")++"\n")
                else done
     if renresFlag then putStrLn ("\nRenamed Resultants: \n"  ++(ppPairs rresultants "=")++"\n")
                   else done
     if postFlag then putStrLn ("\nAfter postunfolding: \n"   ++(ppPairs pprresultants "=")++"\n")
                 else done
     putStrLn ("\nWriting original program into \""++pefile++"\"...\n")
     writeFCY pefile progPE
  where 
     (gr,_)      = gnn_memo prog [] (maxVarIndex e) e                       -- llamada inicial al proceso
     grm         = mark_memo gr
     resultants  = (map delgens (deleteRenamings (extractResultants e grm))) --(extractMainCallLhs prog))
     renaming    = buildRen (map fst resultants) 0
     rresultants = renResultants resultants renaming
     Comb FuncCall f _ = e
     (_,rhs) : _    = rresultants
     prresultants   = postUnfolding rresultants (funsIn rhs)
     pprresultants  = (e,rhs) : prresultants
     fundecls       = map resultant2fundecl pprresultants
     progPE = if mixFlag then fundecl2progA prog fundecls
                         else fundecl2prog prog fundecls
     pefile =      p++"_pe.fcy"

delgens (lhs,rhs) = (lhs, delGEN2 rhs)

gnn_memo :: Prog -> [Expr] -> Int -> Expr -> (GNNTree,[Expr])
gnn_memo prog memo fresh e
 | if pevalDebug 
     then trace ("\nCurrent expression:\n" ++ ppExpr 0 e ++ "\n" ++ "Memo list:\n" ++ concatMap (ppExpr 0) memo ++ "\n") False
     else False                  = (Leaf e,[])                                 --never touched
 | isCons e                      =  (Leaf e, memo)        --se cachan tambien Var y Lit
 | isConsRooted e                =   gnn_memoD prog memo fresh e (decons e)
 | hasGEN e && isOpRooted e      =   gnn_memoG prog memo fresh e (generalize fresh e)
 | otherwise                     =   gnn_memoN prog memo fresh e

gnn_memoD prog memo fresh e es = (TreeD e gres, newmemo)
  where (gres,newmemo) = gnn_memoSeq prog memo fresh es

gnn_memoG prog memo fresh e es = (TreeG e gres, newmemo)
  where (gres, newmemo) = gnn_memoSeq prog memo fresh es

gnn_memoN prog memo fresh e = if b then (LeafM e ne, memo) 
                                   else gnn_memoVar prog memo fresh e
  where (b,ne) = existsRen memo e



                      -- si no es case var, debemos asegurar no GENs en e, puede tratarse de una secuencia
                      -- de case(case(case e)) donde e es un constructor o funcall posiblemente con GENs

gnn_memoVar :: Prog -> [Expr] -> Int -> Expr -> (GNNTree,[Expr])
gnn_memoVar prog memo fresh e = 
  if isCaseVar e then gnn_memoVarBind prog memo fresh e
                 else if (isCase e) then gnn_memoRLNT2 prog memo fresh e 
                          else gnn_memoRLNT prog memo fresh e
          
                      -- aqui llegamos solo cuando es (case exp) y generalizamos (case fun), evitando unfolding
gnn_memoRLNT2 prog memo fresh e =        
  if ((isOpRooted ca) && (hasGEN ca)) then gnn_memoG prog memo fresh e ((changeCaseArg e gca):gcas)               
               else gnn_memoRLNT prog memo fresh e        
  where ca = getcaseArg e
        (gca:gcas) = (generalize (fresh + 10*(maxVarIndex e) + 1) ca)

     

gnn_memoVarBind prog memo fresh e = (TreeN new gres, newmemo)
  where new = propagateBindings e
        (gres,newmemo) = gnn_memoSeq prog memo fresh (extractBranches new)


propagateBindings :: Expr -> Expr
propagateBindings (Case ctype (Var n) ces) = Case ctype (Var n) (propBinding n ces)

gnn_memoRLNT prog memo fresh e = 
  (TreeN e [TreeG exp nnexps], newnewmemo)
  where
    [exp] = unfold prog [] fresh --memo fresh 
                   (if pevalDebugUnfold then (trace ("\nExp to UNFOLD:" ++ (ppExpr 0 e) ++ "\n") e)
                       else e)
    nexps = flattenOC (if pevalDebugUnfold then (trace ("\nEXP: " ++  (ppExpr 0 exp) ++ "\n") exp)
                       else exp)
       --(nfresh,nexps1,nexps2) = flattenSeq fresh 
       --                         (if pevalDebugUnfold then (trace ("\nEXPS: " ++ (ppFormatList (ppExpr 0) [exp]) ++ "\n") [exp])
       --                             else [exp])
    e2 = delGEN2 e
    newmemo = if isOpRooted e2 then e2 : memo else memo
    (nnexps,newnewmemo) = gnn_memoSeq prog newmemo fresh 
                          (if pevalDebugUnfold then (trace ("\nNEXPS: " ++ (ppFormatList (ppExpr 0) nexps) ++ "\n") nexps)
                              else nexps)
    --(nnexps,newnewmemo) = gnn_memoSeq prog newmemo (nfresh) (nexps1++nexps2)
                   --(trace ("\nNEXPS: " ++ (ppFormatList (ppExpr 0) (nexps1++nexps2)) ++ "\n") (nexps1++nexps2))
    --(nnexps,newnewmemo) = unzip (map (gnn_memo prog newmemo (fresh+10)) nexps)


flattenSeq fresh [] = (fresh,[],[])
flattenSeq fresh (e:es) = 
   if nerest==[] then (nnfresh, nes, nrest)
                 else (nnfresh, (head nerest):nes,(tail nerest)++nrest)
  where (nfresh,nerest) = flatten fresh e
                         --(trace ("\nFlattenSeq (e): " ++ (ppExpr 0 e) ++ "\n") e)
        (nnfresh,nes,nrest) = flattenSeq nfresh es

--this function flattens an expression:
flatten fresh (Var e) = (fresh, [])
flatten fresh (Lit l) = (fresh, [])
flatten fresh (Comb ConsCall c es) = (nfresh, nes++nnes)
  where (nfresh,nes,nnes) = flattenSeq fresh es
--concatMap (flatten fresh) es
flatten fresh (Comb FuncCall f es) = (nfresh, (Comb FuncCall f nes) : rest)
  where (nfresh,nes,rest) = flattenSeq_ fresh es
flatten fresh (Case ctype e ces) = (nnfresh, nerest ++ nes ++ nrest)
  where (nfresh,nerest) = flatten fresh e
        (nnfresh,nes,nrest) = flattenSeq nfresh (map branch ces)
--(flatten fresh e) ++ (concatMap (flattenBranch fresh) ces)

branch (Branch p e) = e
--flattenBranch fresh (Branch p e) = flatten fresh e

flattenSeq_ fresh [] = (fresh,[],[])
flattenSeq_ fresh (e:es) = (nnfresh,ne:nes,rest++nrest)
  where (nfresh,ne:rest) = flatten_ fresh e 
        (nnfresh,nes,nrest) = flattenSeq_ nfresh es

flatten_ fresh (Var e) = (fresh, [Var e])
flatten_ fresh (Lit l) = (fresh, [Lit l])
flatten_ fresh (Comb ConsCall c es) = (nfresh, (Comb ConsCall c nes) : rest)
  where (nfresh,nes,rest) = flattenSeq_ fresh es
flatten_ fresh (Comb FuncCall f es) = 
  (nfresh, (Var (fresh + 10*(maxVarIndex (Comb FuncCall f es)) + 1)) : nes)
  where (nfresh,nes) = flatten (fresh+1) (Comb FuncCall f es)
--  where (nfresh,nes,rest) = flattenSeq (fresh+1) es

--this function implements the sequential traversal of arguments in
--a function or constructor call
--it returns the GNN trees for each argument together with the final
--memo list
gnn_memoSeq :: Prog -> [Expr] -> Int -> [Expr] -> ([GNNTree],[Expr])
gnn_memoSeq prog memo fresh [] = ([],memo)
gnn_memoSeq prog memo fresh (e:es) = (gr : gres, lastmemo)
  where (gr,newmemo) = gnn_memo prog memo fresh e
        (gres,lastmemo) = gnn_memoSeq prog newmemo fresh es


--this function flattens an expression (only deletes cases):
flattenOC (Var e) = []
flattenOC (Lit l) = []
flattenOC (Comb ConsCall c es) = [Comb ConsCall c es]
flattenOC (Comb FuncCall f es) = [Comb FuncCall f es]
flattenOC (Case ctype e ces) = ne ++ nces
  where ne = flattenOC e
        nces = concatMap flattenOC (map branch ces)

-- get case argument

getcaseArg :: Expr -> Expr
getcaseArg e =  
  case e of
    Case _ ex _ -> getcaseArg ex  -- si es un case anidado busca el arg. mas interno
    _           -> e              -- si no es case ya, devolver ex 


-- changeCaseArgument

changeCaseArg :: Expr -> Expr -> Expr
changeCaseArg e gene =  
  case e of
    Case ctype ex ces -> Case ctype (changeCaseArg ex gene) ces  -- llegara al arg mas interno
    _                 -> gene                                    -- devuleve arg. generalizado 



--checks if an expression is a case structure 
isCase :: Expr -> Bool
isCase e = case e of
                Case _ _ _ -> True
                _ -> False


