------------------------------------------------------------------------------
--- This module implements the post-unfolding simplification process
---
--- @author JGuadalupe Ramos, Josep Silva, German Vidal
--- @version June 2005
------------------------------------------------------------------------------

module PostUnf(postUnfolding) where

import FlatCurry

import Utils(vars2index,eB,substitute)
import RLNT(propBinding)

postUnfolding :: [(Expr,Expr)] -> [QName] -> [(Expr,Expr)]
postUnfolding prog funcs =
  if newProg==prog then prog
     else postUnfolding newProg funcs 
     where newProgTemp = postUnfoldingAux prog prog
           newProgAux = removeRedundantRules newProgTemp newProgTemp funcs
           newProg = removeFails newProgAux

postUnfoldingAux :: [(Expr,Expr)] -> [(Expr,Expr)] -> [(Expr,Expr)]
postUnfoldingAux _ [] = []
postUnfoldingAux prog ((l,r):es) =
  if r==rUnf then (l,r) : (postUnfoldingAux prog es)
             else (l,rUnf) : es --DO NOT APPLY IT TO 'es'..
       where rUnf = try_unfold prog r

try_unfold :: [(Expr,Expr)] -> Expr -> Expr
try_unfold _ (Var x) = Var x
try_unfold _ (Lit l) = Lit l
try_unfold prog (Comb ConsCall c es) = Comb ConsCall c (try_unfoldL prog es)
--try_unfold prog (Comb FuncPartCall c es) = Comb FuncPartCall c (try_unfoldL prog es)
try_unfold prog (Comb FuncCall f es) = 
  if nonRecursive prog f
     then unfoldFC prog prog (Comb FuncCall f es)
     else if uselessFunc prog f
             then unfoldFC prog prog (Comb FuncCall f es)
             else Comb FuncCall f (try_unfoldL prog es)
try_unfold prog (Free gc e) = Free gc (try_unfold prog e)
try_unfold prog (Case ctype e cases) = 
  Case ctype (try_unfold prog e) (map (try_unf_cases prog) cases)
try_unfold prog (Or e1 e2) = 
  Or (try_unfold prog e1) (try_unfold prog e2)

try_unfoldL :: [(Expr,Expr)] -> [Expr] -> [Expr]
try_unfoldL _ [] = []
try_unfoldL prog (e:es) = (try_unfold prog e) : (try_unfoldL prog es)

try_unf_cases :: [(Expr,Expr)] -> BranchExpr -> BranchExpr
try_unf_cases prog (Branch p e) = Branch p (try_unfold prog e) 

-----------

removeRedundantRules :: [(Expr,Expr)] -> [(Expr,Expr)] -> [QName] -> [(Expr,Expr)]
removeRedundantRules [] _ _ = []
removeRedundantRules (rule:es) prog funcs =
  if (funcRule rule) `elem` funcs 
     then rule : (removeRedundantRules es prog funcs)
     else if (countFuncCalls prog (funcRule rule))==0
          then removeRedundantRules es prog funcs
          else rule : (removeRedundantRules es prog funcs)

funcRule ((Comb FuncCall f _),_) = f

------------

removeFails :: [(Expr,Expr)] -> [(Expr,Expr)]
removeFails [] = []
removeFails ((lhs,rhs):es) = (lhs,(removeFB rhs)) : (removeFails es)

removeFB :: Expr -> Expr
removeFB x = case x of
 (Case ctype e ces) -> let newces = removeFailedBranches ces in
                       if newces==[] 
                       then Comb FuncCall ("Prelude","failed") [] 
                       else Case ctype (removeFB e) (map removeFBbranch newces)
 (Comb c f es) -> Comb c f (map removeFB es)
 (Free v e) -> Free v (removeFB e)
 (Or a b) -> Or (removeFB a) (removeFB b)
 e -> e

----

nonRecursive :: [(Expr,Expr)] -> QName -> Bool
nonRecursive [] _ = False
nonRecursive ((Comb FuncCall g _, r) : rules) f =
  if f==g then (countFuncs r)==0     --only rhs's with ctes and vbles are allowed.. TO DO..
          else nonRecursive rules f

--note that countFuncs counts all defined functions and countFuncCalls only the
--occurrences of a given function..
countFuncs :: Expr -> Int
countFuncs (Var _) = 0
countFuncs (Lit _) = 0
countFuncs (Comb t f es) 
  | t==ConsCall             = countFuncsL es
  | f==("Prelude","failed") = 0
  | otherwise               = 1  -- + (countFuncsL es)  --not needed for its current use..
countFuncs (Free _ e) = countFuncs e
countFuncs (Case _ e cases) = 
  (countFuncs e) + (countFuncsL (eB cases))
countFuncs (Or e1 e2) = (countFuncs e1) + (countFuncs e2)

countFuncsL :: [Expr] -> Int
countFuncsL [] = 0
countFuncsL (e:es) =
   (countFuncs e) + (countFuncsL es)

------------------

unfoldFC :: [(Expr,Expr)] -> [(Expr,Expr)] -> Expr -> Expr
unfoldFC [] prog (Comb FuncCall f es) = Comb FuncCall f (try_unfoldL prog es)
unfoldFC (((Comb FuncCall g vars),r):rules) prog (Comb FuncCall f es) = 
  if f==g && n==0 then propCaseVars (substitute (vars2index vars) es r)
                  else unfoldFC rules prog (Comb FuncCall f es)
  where n = countFuncCalls [((Comb FuncCall g vars),r)] g

propCaseVars :: Expr -> Expr
propCaseVars e = case e of
  (Case ctype (Var x) ces) ->
              Case ctype (Var x) (map propCaseVarsBranch (propBinding x ces))
  _ -> e
propCaseVarsBranch :: BranchExpr -> BranchExpr
propCaseVarsBranch (Branch pat e) = Branch pat (propCaseVars e)

----------------

uselessFunc :: [(Expr,Expr)] -> QName -> Bool
uselessFunc prog f = (countFuncCalls prog f)==1

countFuncCalls :: [(Expr,Expr)] -> QName -> Int
countFuncCalls [] _ = 0
countFuncCalls ((_,r):es) f = (countFC r f) + (countFuncCalls es f)

countFC :: Expr -> QName -> Int
countFC (Var _) _ = 0
countFC (Lit _) _ = 0
--countFC (Comb ConsCall _ es) f = countFCList es f
--countFC (Comb FuncPartCall _ es) f = countFCList es f
countFC (Comb _ g es) f = 
  if f==g then 1 + (countFCList es f) else countFCList es f
countFC (Free _ e) f = countFC e f
countFC (Case _ e cases) f = 
  (countFC e f) + (countFCList (eB cases) f)
countFC (Or e1 e2) f = (countFC e1 f) + (countFC e2 f)

countFCList [] _ = 0
countFCList (e:es) f =
   (countFC e f) + (countFCList es f)

-------------

removeFailedBranches :: [BranchExpr] -> [BranchExpr]
removeFailedBranches [] = []
removeFailedBranches ((Branch pat e):es) = 
  if e==(Comb FuncCall ("Prelude","failed") []) 
     then removeFailedBranches es
     else (Branch pat e) : (removeFailedBranches es)

removeFBbranch :: BranchExpr -> BranchExpr
removeFBbranch (Branch pat e) = Branch pat (removeFB e)


----------------

