------------------------------------------------------------------------------
--- This module implements the extraction of resultants from a 
--- GNN tree 
--- 
--- @author JGuadalupe Ramos, Josep Silva, German Vidal
--- @version June 2005
------------------------------------------------------------------------------

module Resultants(extractResultants,buildRen,renResultants) where

import FlatCurry
import Unsafe
import List

import OffPretty(ppExpr,ppTree)
import GNNTrees(GNNTree(..))
import General(delGEN)
import Utils(isCaseVar,nonLocalVarsInExp,existsRen,searchConstrInstance,
             subVars,subTerms,searchInstance,Subst(..),index2vars,substitute)

--extracts the resultants from a GNN tree and an initial call:
extractResultants :: Expr -> GNNTree -> [(Expr,Expr)]
extractResultants lhs grm 
  --| trace ("\nCurrent expression:\n" ++ ppExpr 0 lhs ++ "\nwith TREE:\n"++ppTree 0 grm ++"\n") True
  = case grm of
   Leaf e -> [(lhs, e)]
   LeafM e _ -> [(lhs, e)]
   TreeG e grs -> (lhs, delGEN e) : concatMap eR grs --if lhs==e then concat (map eR grs)
                                                     --else (lhs, delGEN e) : concat (map eR grs)
   TreeD e grs -> extractResD lhs e grs --if lhs==e then concat (map eR grs)
                                        --else (lhs, delGEN e) : concat (map eR grs) 
   TreeN e grs -> if isCaseVar e then eRcaseVar e grs lhs
                                 else concatMap (extractResultants lhs) grs
   TreeNM e grs -> if isCaseVar e then eRcaseVar e grs lhs
                                  else (lhs, delGEN e) : concatMap (extractResultants (delGEN e)) grs

extractResD lhs (Comb ConsCall f es) grs = (lhs, Comb ConsCall f nes) : (concat rest)
  where resultants = map (extractResultants lhs) grs
        (nes,rest) = extractArgs es resultants

--extracts the resultants associated to the arguments of a constructor application:
extractArgs [] [] = ([],[])
extractArgs (_:es) (((_,rhs):other):rest) = (rhs : rhsl, other : otherl)
  where (rhsl, otherl) = extractArgs es rest

--extracts resultants from a GNN tree:
eR gr = case gr of
          TreeG _ grs -> concatMap eR grs
          TreeD _ grs -> concatMap eR grs
          TreeN e grs -> concatMap (extractResultants e)  grs
          TreeNM e grs -> concatMap (extractResultants e) grs
          _ -> [] --for Leaf and LeafM

eRcaseVar (Case ctype v ces) grs lhs = 
     (lhs, Case ctype v nces) : rest
  where 
    (nces,rest) = eRBranches ces grs 


eRBranches [] [] = ([],[])
eRBranches (Branch p e : ces) (gre:gres) = (Branch p ne : nces, res++nres)
  where (_,ne) : res = extractResultants e gre
        (nces, nres) = eRBranches ces gres

-- independent renaming of a set of terms:
-- return tuples (e,e') where e' is the independent renaming of e

buildRen :: [Expr] -> Int -> [(Expr,Expr)]
buildRen [] _ = []

buildRen (Comb ftype (mod,fname) args : es) n =
  (Comb ftype (mod,fname) args, 
   Comb FuncCall (mod,fname++"_pe"++(show n)) (index2vars vars)) 
  : (buildRen es (n+1))
 where vars = nub (nonLocalVarsInExp (Comb ftype (mod,fname) args))

buildRen (Case ctype e cases : es) n = 
   (Case ctype e cases,
    Comb FuncCall ("Prelude","case_pe" ++ (show n)) (index2vars vars))
   : (buildRen es (n+1))
 where vars = nub (nonLocalVarsInExp (Case ctype e cases) )

buildRen (Free v c : es) n =
   (Free v c,
    Comb FuncCall ("Prelude","free_pe" ++ (show n)) (index2vars vars))
   : (buildRen es (n+1))
 where vars = nub (nonLocalVarsInExp (Free v c)) 

buildRen (Or e1 e2 : es) n =
   (Or e1 e2,
    Comb FuncCall ("Prelude","or_pe" ++ (show n)) (index2vars vars))
   : (buildRen es (n+1))
 where vars = nub (nonLocalVarsInExp (Or e1 e2)) 

--renames resultants w.r.t. an independent renaming:
renResultants :: [(Expr,Expr)] -> [(Expr,Expr)] -> [(Expr,Expr)]
renResultants [] _ = []
renResultants ((lhs,rhs) : res) renaming =
  (renLHS lhs renaming, renRHS renaming rhs) : renResultants res renaming

renLHS :: Expr -> [(Expr,Expr)] -> Expr
renLHS e ((s,ns):rest) = if b then ns else renLHS e rest
  where (b,_) = existsRen [s] e

-- renaming function for arbitrary expressions: ren_rho
-- rename of the RHS according to the independent renaming rho

-- TODO: it can be implemented better by using the information
--       from the GNNTree (i.e., the backwards arrows), e.g.,
--       one can leave the GEN functions in the GNN Tree so that
--       renaming proceeds by looking for variants of each
--       subexpression..

renRHS  :: [(Expr,Expr)] -> Expr -> Expr
renRHS _ (Var x) = Var x
renRHS _ (Lit x)  = Lit x
renRHS renaming (Comb FuncCall c e) = renFun renaming (Comb FuncCall c e)
renRHS renaming (Comb ConsCall c e) = Comb ConsCall c (map (renRHS renaming) e)
renRHS renaming (Free vars e) = Free vars (renRHS renaming e) 
renRHS renaming (Case ctype e cases) = 
  Case ctype (renRHS renaming e) (map (renCase renaming) cases)
renRHS renaming (Or e1 e2) = Or (renRHS renaming e1) (renRHS renaming e2)

renCase :: [(Expr,Expr)] -> BranchExpr -> BranchExpr
renCase renaming (Branch p e) = Branch p (renRHS renaming e) 

renFun :: [(Expr,Expr)] -> Expr -> Expr
renFun renaming e = 
  if sub==FSub
     then renFun_ renaming e
     else substitute (subVars sub) (subTerms sub) exp
  where (sub, exp) = searchConstrInstance e renaming

renFun_ :: [(Expr,Expr)] -> Expr -> Expr
renFun_ renaming e = 
  if subs==FSub
     then renRHS renaming e
     else substitute (subVars subs) (map (renRHS renaming) (subTerms subs)) e2
  where (subs, e2) = searchInstance e renaming

