------------------------------------------------------------------------------
--- This module implements several useful utilities 
--- 
--- @author JGuadalupe Ramos, Josep Silva, German Vidal
--- @version June 2005
------------------------------------------------------------------------------

module Utils(maxVarIndex,triple1st,triple2nd,substitute,
             substituteAll,index2vars,vars2index,getFuncDecls,maxList,
             isCaseVar,nonLocalVarsInExp,existsRen,searchConstrInstance,
             subVars,subTerms,searchInstance,Subst(..),eB,extractMainCallLhs,extractMainCallRhs,
             deleteRenamings,funsIn,isCons,isConsRooted,decons,isOpRooted,
             extractBranches,isFlatVar) where

import FlatCurry
import Unsafe

--extracts the rhs of the main (i.e., the first) call of a program:
extractMainCallRhs :: Prog -> Expr
extractMainCallRhs (Prog _ _ _ (Func _ _ _ _ (Rule _ e) : _) _) = e

--extracts the lhs of the main (i.e., the first) call of a program:
extractMainCallLhs :: Prog -> Expr
extractMainCallLhs (Prog _ _ _ (Func fname arity _ _ _ : _) _) = 
  Comb FuncCall fname (makeListVars 0 (arity-1))

--builds a list of variable arguments of a given arity:
makeListVars :: Int -> Int -> [Expr]
makeListVars n m = if n==m then [Var n] else Var n : makeListVars (n+1) m

--getting function declarations from a program
getFuncDecls (Prog _ _ _ funs _) = funs

--checks if a expression is a case structure with a vble argument
isCaseVar :: Expr -> Bool
isCaseVar e = case e of
                Case _ (Var _) _ -> True
                _ -> False

--checks if a expression is an operation-rooted term
isOpRooted :: Expr -> Bool
isOpRooted e = case e of
                Comb FuncCall _ _ -> True
                _ -> False

--extracts the branches of a case expression
extractBranches :: Expr -> [Expr]
extractBranches (Case _ _ ces) = eB ces
eB [] = []
eB (Branch _ e : ces) = e : eB ces


--generic deconstruction: 
decons :: Expr -> [Expr]
decons (Comb ConsCall _ es) = es

--is it a constructor term?
isCons :: Expr -> Bool
isCons (Var _) = True
isCons (Lit _) = True
isCons (Comb ftype _ es) = 
    case ftype of
         ConsCall -> and (map isCons es)
         _ -> False
isCons (Free _ e) = isCons e
isCons (Or e1 e2) = (isCons e1) && (isCons e2)
isCons (Case _ e ces) = False  --(isCons e) && (and (map isConsBranch ces))

isConsBranch (Branch _ e) = isCons e

--is it a constructor-rooted term?
isConsRooted :: Expr -> Bool
isConsRooted e = case e of
                      Comb ConsCall _ _ -> True
                      _ -> False

--is it a variable?
isFlatVar e = case e of
                 Var _ -> True
                 _ -> False

--is it a constructor substitution?
isConstructor (Sub _ expr) = and (map isCons expr)


--for extracting triple elements:
triple1st (a,_,_) = a
triple2nd (_,b,_) = b
triple3rd (_,_,c) = c

--converts an integer list into a list of vars 
index2vars :: [VarIndex] -> [Expr]
index2vars = map (\i->Var i)

--and vice versa:
vars2index :: [Expr] -> [VarIndex]
vars2index = map (\(Var i)->i)

-- get the maximum index of all variables in an expression
-- (or -1 if there is no variable)
maxVarIndex :: Expr -> Int
maxVarIndex (Var v)         = v
maxVarIndex (Lit _)         = -1
maxVarIndex (Comb _ _ exps) = maxList (map maxVarIndex exps)
maxVarIndex (Free vars e) = max (maxList vars) (maxVarIndex e)
maxVarIndex (Or e1 e2)      = max (maxVarIndex e1) (maxVarIndex e2)
maxVarIndex (Case _ ce cs)   = max (maxVarIndex ce) (maxList (map maxCase cs))
  where maxCase (Branch (Pattern _ pvs) e) = max (maxList pvs) (maxVarIndex e)
        maxCase (Branch (LPattern _)   e) = maxVarIndex e

-- maximum of two numbers:   *** now defined in the prelude ***
--max i j = if i<=j then j else i

-- maximum of a list of naturals (-1 if list is empty)
maxList xs = foldr max (-1) xs

--check if a term (second argument) appears (modulo renaming) 
--in a list (first argument)
existsRen :: [Expr] -> Expr -> (Bool,Expr)

existsRen [] _ = (False, Var (-1))
existsRen (e1:es) e = 
  if  not (instance e1 e == FSub) &&  
      not (instance e e1 == FSub)
                      then (True,e1)
                      else existsRen es e

-- data type for substitutions:
data Subst = Sub [Int] [Expr] | FSub

-- check whether one expression is an instance of another and compute
-- the corresponding substitution:
instance :: Expr -> Expr -> Subst
instance a b =
  if a==b then Sub [] []
          else if isFlatVar b 
                  then instanceVar a b
                  else if headExpr a == headExpr b
                          then instanceS a b
                          else FSub

instanceVar :: Expr -> Expr -> Subst
instanceVar a (Var x) = Sub [x] [a]

--no need to treat vbles (otherwise we don't call to instanceS..)
instanceS :: Expr -> Expr -> Subst
instanceS (Lit x) (Lit y) = 
  if x==y then Sub [] [] else FSub
instanceS (Comb c1 f1 e1) (Comb c2 f2 e2) =
  if c1==c2 && f1==f2 then instanceL e1 e2 (Sub [] [])
                      else FSub
instanceS (Case c1 e1 ces1) (Case c2 e2 ces2) =
  if c1==c2 then instanceCase ces1 ces2 (instance e1 e2)
            else FSub
instanceS (Free _ e1) (Free _ e2) = instance e1 e2
instanceS (Or e1 e2) (Or f1 f2) = instanceL [e1,e2] [f1,f2] (Sub [] [])

instanceL :: [Expr] -> [Expr] -> Subst -> Subst
instanceL [] [] s = s
instanceL (e1:es1) (e2:es2) subs 
  | newsub == FSub    =  FSub
  | clash newsub subs =  FSub
  | otherwise         =  instanceL es1 es2 (composeSubs subs newsub)
  where newsub = instance e1 e2

instanceCase :: [BranchExpr] -> [BranchExpr] -> Subst -> Subst
instanceCase _ _ FSub = FSub
instanceCase ces1 ces2 (Sub sx vx) =
  instanceCaseL ces1 ces2 (Sub sx vx)

instanceCaseL :: [BranchExpr] -> [BranchExpr] -> Subst -> Subst
instanceCaseL [] [] s = s
instanceCaseL (e1:es1) (e2:es2) subs 
  | newsub == FSub    =  FSub
  | clash newsub subs =  FSub
  | otherwise         =  instanceCaseL es1 es2 (composeSubs subs newsub)
  where newsub = instanceCaseBranch e1 e2

instanceCaseBranch :: BranchExpr -> BranchExpr -> Subst
instanceCaseBranch (Branch (Pattern c1 _) e1) (Branch (Pattern c2 _) e2) = 
  if c1 == c2 then instance e1 e2
              else FSub
instanceCaseBranch (Branch (LPattern c1) e1) (Branch (LPattern c2) e2) = 
  if c1 == c2 then instance e1 e2
              else FSub

--returns the name of the function or the type of expression:
headExpr :: Expr -> String
headExpr (Var _) = "Var"
headExpr (Lit _) = "Lit"
headExpr (Comb _ (m,f) _) = m++f
headExpr (Case Rigid _ _) = "Rigid"
headExpr (Case Flex _ _) = "Flex"
headExpr (Free _ _) = "Constr"
headExpr (Or _ _) = "Or"

-- (substitute vars exps expr) = expr[vars/exps]
-- i.e., replace all occurrences of vars by corresponding exps in the
-- expression expr
substitute :: [Int] -> [Expr] -> Expr -> Expr
substitute vars exps expr = substituteAll vars exps 0 expr

-- (substituteAll vars exps base expr):
-- substitute all occurrences of variables by corresponding expressions:
-- * substitute all occurrences of var_i by exp_i in expr
--   (if vars=[var_1,...,var_n] and exps=[exp_1,...,exp_n])
-- * substitute all other variables (Var j) by (Var (base+j))
--
-- here we assume that the new variables in guards and case patterns
-- do not occur in the list "vars" of replaced variables!

substituteAll :: [Int] -> [Expr] -> Int -> Expr -> Expr
substituteAll vars exps b (Var i) = replaceVar vars exps i
  where replaceVar [] [] var = Var (b+var)
        replaceVar (v:vs) (e:es) var = if v==var then e
                                                 else replaceVar vs es var
substituteAll _  _  _ (Lit l) = Lit l
substituteAll vs es b (Comb combtype c exps) =
                 Comb combtype c (map (substituteAll vs es b) exps)
substituteAll vs es b (Free vars e) =
                 Free (map (+b) vars) (substituteAll vs es b e)
substituteAll vs es b (Or e1 e2) =
                 Or (substituteAll vs es b e1) (substituteAll vs es b e2)
substituteAll vs es b (Case ctype e cases) =
   Case ctype (substituteAll vs es b e) (map (substituteAllCase vs es b) cases)

substituteAllCase :: [Int] -> [Expr] -> Int -> BranchExpr -> BranchExpr
substituteAllCase vs es b (Branch (Pattern l pvs) e) =
                 Branch (Pattern l (map (+b) pvs)) (substituteAll vs es b e)
substituteAllCase vs es b (Branch (LPattern l) e) =
                 Branch (LPattern l) (substituteAll vs es b e)

-- composition of substitutions:

composeSubs :: Subst -> Subst -> Subst
composeSubs (Sub vars1 exps1) (Sub vars2 exps2) = 
  (Sub (vars1 ++ vars2) (newExps ++ exps2))
  where newExps = (map (substitute vars2 exps2) exps1)
composeSubs FSub (Sub _ _) = FSub
composeSubs (Sub _ _) FSub = FSub
composeSubs FSub FSub = FSub

--checks whether there is a clash of variables:
clash :: Subst -> Subst -> Bool
clash (Sub vs es) sub2 = clash_ vs es sub2

clash_ [] [] _ = False
clash_ (v:vs) (e:es) (Sub vs2 es2) =
  if ve==Var (-1) then clash_ vs es (Sub vs2 es2) 
                   else not (e==ve)
  where ve = findVal v vs2 es2

--returns the value associated to a variable in a substitution:
findVal :: Int -> [Int] -> [Expr] -> Expr
findVal _ [] [] = Var (-1)
findVal w (v:vs) (e:es) = if w==v then e else findVal w vs es

--returns the non-local variables of an expression

nonLocalVarsInExp :: Expr -> [Int]

nonLocalVarsInExp (Var v) = [v]
nonLocalVarsInExp (Lit _) = []
nonLocalVarsInExp (Comb _ _ exps) = concat (map nonLocalVarsInExp exps)
nonLocalVarsInExp (Free vars e) = delArgs vars (nonLocalVarsInExp e)
nonLocalVarsInExp (Or e1 e2) = nonLocalVarsInExp e1 ++ nonLocalVarsInExp e2
nonLocalVarsInExp (Case _ e cases) = nonLocalVarsInExp e ++ concat (map varsNonLocal cases)


--returns the non-local variables of the different scrutinees of 
--a case branch
varsNonLocal :: BranchExpr -> [Int] 
varsNonLocal (Branch (Pattern _ as) e) = delArgs as (nonLocalVarsInExp e) 
varsNonLocal (Branch (LPattern _) e)   = nonLocalVarsInExp e

-- delete argument variables in a list of variables
delArgs :: [a] -> [a] -> [a]
delArgs [] vs = vs
delArgs (x:xs) vs = delArgs xs (deleteAll x vs)

-- delete all occurrences of an element in a list:
deleteAll          :: a -> [a] -> [a]
deleteAll _ []     = []
deleteAll x (y:ys) = if x==y then deleteAll x ys else y : deleteAll x ys

--extract vars and terms of a given (nonfailed) substitution
subVars :: Subst -> [Int]
subVars (Sub v _) = v
subVars FSub = trace "\n\n ERROR!!!!\n\n" [-2]
subTerms :: Subst -> [Expr]
subTerms (Sub _ e) = e
subTerms FSub = trace "\n\n ERROR!!!\n\n" [Var (-2)]

--returns the matching substitution between an expression and 
--a set of expressions (FSbub otherwise), 
--as well as the matched term..
search_instance :: Expr -> [Expr] -> (Subst,Expr)
search_instance _ [] = (FSub, Var (-1))
search_instance e (e1:es) = 
  if sub==FSub then search_instance e es
               else (sub,e)
  where sub = instance e e1

--checks whether a term is an instance of the renaming set.. 
--returns both the substitution and the selected term.. 
searchInstance :: Expr -> [(Expr,Expr)] -> (Subst,Expr)
searchInstance _ [] = (FSub, Var (-1))
searchInstance e ((e1,e2):es) =
  if sub /= FSub then (sub, e2)
                 else searchInstance e es
  where sub = instance e e1

--checks whether a term is a constructor instance of the renaming set.. 
--returns the selected term of the renaming set only if the substitution 
--is 'constructor' 
--(i.e., the matching substitution does not contain defined function symbols..)
searchConstrInstance :: Expr -> [(Expr,Expr)] -> (Subst,Expr)
searchConstrInstance e [] = (FSub,e)
searchConstrInstance e ((e1,e2):es) = 
  if (sub /= FSub) && (isConstructor sub)
                         then (sub,e2)
                         else searchConstrInstance e es
  where sub = instance e e1

--returns a list with the function names (qualified) in an expression:
funsIn :: Expr -> [QName]
funsIn (Var _) = []
funsIn (Lit _) = []
funsIn (Comb ftype fname es) = 
  case ftype of
    FuncCall -> fname : concat (map funsIn es)
    _ -> concat (map funsIn es)
funsIn (Free _ e) = funsIn e
funsIn (Or e1 e2) = (funsIn e1) ++ (funsIn e2)
funsIn (Case _ e ces) = (funsIn e) ++ (concat (map funsInBranch ces))

funsInBranch (Branch _ e) = funsIn e

--remove duplicated expressions (modulo variable renaming) 
deleteRenamings :: [(Expr,Expr)] -> [(Expr,Expr)]
deleteRenamings [] = []
deleteRenamings ((lhs,rhs):es) = 
     if b then deleteRenamings es
          else (lhs,rhs) : deleteRenamings es
  where (b,_) = existsRen (map fst es) lhs
